"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedBlackTree = exports.RedBlackTreeNode = void 0;
const bst_1 = require("./bst");
class RedBlackTreeNode extends bst_1.BSTNode {
    /**
     * The constructor function initializes a Red-Black Tree Node with a key, an optional value, and a
     * color.
     * @param {K} key - The key parameter is of type K and represents the key of the node in the
     * Red-Black Tree.
     * @param {V} [value] - The `value` parameter is an optional parameter that represents the value
     * associated with the key in the Red-Black Tree Node. It is not required and can be omitted when
     * creating a new instance of the Red-Black Tree Node.
     * @param {RBTNColor} color - The `color` parameter is used to specify the color of the Red-Black
     * Tree Node. It is an optional parameter with a default value of `'BLACK'`.
     */
    constructor(key, value, color = 'BLACK') {
        super(key, value);
        this._color = color;
    }
    /**
     * The function returns the color value of a variable.
     * @returns The color value stored in the private variable `_color`.
     */
    get color() {
        return this._color;
    }
    /**
     * The function sets the color property to the specified value.
     * @param {RBTNColor} value - The value parameter is of type RBTNColor.
     */
    set color(value) {
        this._color = value;
    }
}
exports.RedBlackTreeNode = RedBlackTreeNode;
class RedBlackTree extends bst_1.BST {
    /**
     * This is the constructor function for a Red-Black Tree data structure in TypeScript.
     * @param keysOrNodesOrEntries - The `keysOrNodesOrEntries` parameter is an iterable object that can
     * contain keys, nodes, or entries. It is used to initialize the RBTree with the provided keys,
     * nodes, or entries.
     * @param [options] - The `options` parameter is an optional object that can be passed to the
     * constructor. It allows you to customize the behavior of the RBTree. It can include properties such
     * as `compareKeys`, `compareValues`, `allowDuplicates`, etc. These properties define how the RBTree
     * should compare keys and
     */
    constructor(keysOrNodesOrEntries = [], options) {
        super([], options);
        this._root = this.NIL;
        if (keysOrNodesOrEntries) {
            this.addMany(keysOrNodesOrEntries);
        }
    }
    /**
     * The function returns the root node of a tree or undefined if there is no root.
     * @returns The root node of the tree structure, or undefined if there is no root node.
     */
    get root() {
        return this._root;
    }
    /**
     * The function creates a new Red-Black Tree node with the specified key, value, and color.
     * @param {K} key - The key parameter represents the key of the node being created. It is of type K,
     * which is a generic type representing the key's data type.
     * @param {V} [value] - The `value` parameter is an optional parameter that represents the value
     * associated with the key in the node. It is not required and can be omitted if not needed.
     * @param {RBTNColor} color - The "color" parameter is used to specify the color of the node in a
     * Red-Black Tree. It is an optional parameter with a default value of "'BLACK'". The color
     * can be either "'RED'" or "'BLACK'".
     * @returns The method is returning a new instance of a RedBlackTreeNode with the specified key,
     * value, and color.
     */
    createNode(key, value, color = 'BLACK') {
        return new RedBlackTreeNode(key, value, color);
    }
    /**
     * The function creates a Red-Black Tree with the given options and returns it.
     * @param [options] - The `options` parameter is an optional object that contains configuration
     * options for creating the Red-Black Tree. It is of type `RBTreeOptions<K>`, where `K` represents
     * the type of keys in the tree.
     * @returns a new instance of a RedBlackTree object.
     */
    createTree(options) {
        return new RedBlackTree([], Object.assign({ iterationType: this.iterationType }, options));
    }
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `keyValueOrEntryToNode` takes a key, value, or entry and returns a node if it is
     * valid, otherwise it returns undefined.
     * @param {KeyOrNodeOrEntry<K, V, NODE>} keyOrNodeOrEntry - The key, value, or entry to convert.
     * @param {V} [value] - The value associated with the key (if `keyOrNodeOrEntry` is a key).
     * @returns {NODE | undefined} - The corresponding Red-Black Tree node, or `undefined` if conversion fails.
     */
    keyValueOrEntryToNode(keyOrNodeOrEntry, value) {
        let node;
        if (keyOrNodeOrEntry === null || keyOrNodeOrEntry === undefined) {
            return;
        }
        else if (this.isNode(keyOrNodeOrEntry)) {
            node = keyOrNodeOrEntry;
        }
        else if (this.isEntry(keyOrNodeOrEntry)) {
            const [key, value] = keyOrNodeOrEntry;
            if (key === undefined || key === null) {
                return;
            }
            else {
                node = this.createNode(key, value, 'RED');
            }
        }
        else if (!this.isNode(keyOrNodeOrEntry)) {
            node = this.createNode(keyOrNodeOrEntry, value, 'RED');
        }
        else {
            return;
        }
        return node;
    }
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     * /
  
     /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function checks if the input is an instance of the RedBlackTreeNode class.
     * @param {KeyOrNodeOrEntry<K, V, NODE>} keyOrNodeOrEntry - The object to check.
     * @returns {boolean} - `true` if the object is a Red-Black Tree node, `false` otherwise.
     */
    isNode(keyOrNodeOrEntry) {
        return keyOrNodeOrEntry instanceof RedBlackTreeNode;
    }
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The "clear" function sets the root node of a data structure to a sentinel value and resets the
     * size counter to zero.
     */
    clear() {
        super.clear();
        this._root = this.NIL;
    }
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The function adds a new node to a Red-Black Tree data structure and returns a boolean indicating
     * whether the operation was successful.
     * @param keyOrNodeOrEntry - The `keyOrNodeOrEntry` parameter can be either a key, a node, or an
     * entry.
     * @param {V} [value] - The `value` parameter is the value associated with the key that is being
     * added to the tree.
     * @returns The method is returning a boolean value. It returns true if the node was successfully
     * added or updated, and false otherwise.
     */
    add(keyOrNodeOrEntry, value) {
        const newNode = this.keyValueOrEntryToNode(keyOrNodeOrEntry, value);
        if (!this.isRealNode(newNode))
            return false;
        const insertStatus = this._insert(newNode);
        if (insertStatus === 'CREATED') {
            // Ensure the root is black
            if (this.isRealNode(this._root)) {
                this._root.color = 'BLACK';
            }
            else {
                return false;
            }
            this._size++;
            return true;
        }
        else
            return insertStatus === 'UPDATED';
    }
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The function `delete` in a binary tree class deletes a node from the tree and fixes the tree if
     * necessary.
     * @param {ReturnType<C> | null | undefined} identifier - The `identifier` parameter is the
     * identifier of the node that needs to be deleted from the binary tree. It can be of any type that
     * is returned by the callback function `C`. It can also be `null` or `undefined` if the node to be
     * deleted is not found.
     * @param {C} callback - The `callback` parameter is a function that is used to retrieve a node from
     * the binary tree based on its identifier. It is an optional parameter and if not provided, the
     * `_DEFAULT_CALLBACK` function is used as the default callback. The callback function should
     * return the identifier of the node to
     * @returns an array of BinaryTreeDeleteResult<NODE> objects.
     */
    delete(identifier, callback = this._DEFAULT_CALLBACK) {
        if (identifier === null)
            return [];
        const results = [];
        const nodeToDelete = this.isRealNode(identifier) ? identifier : this.getNode(identifier, callback);
        if (!nodeToDelete) {
            return results;
        }
        let originalColor = nodeToDelete.color;
        let replacementNode;
        if (!this.isRealNode(nodeToDelete.left)) {
            replacementNode = nodeToDelete.right;
            this._transplant(nodeToDelete, nodeToDelete.right);
        }
        else if (!this.isRealNode(nodeToDelete.right)) {
            replacementNode = nodeToDelete.left;
            this._transplant(nodeToDelete, nodeToDelete.left);
        }
        else {
            const successor = this.getLeftMost(nodeToDelete.right);
            if (successor) {
                originalColor = successor.color;
                replacementNode = successor.right;
                if (successor.parent === nodeToDelete) {
                    if (this.isRealNode(replacementNode)) {
                        replacementNode.parent = successor;
                    }
                }
                else {
                    this._transplant(successor, successor.right);
                    successor.right = nodeToDelete.right;
                    if (this.isRealNode(successor.right)) {
                        successor.right.parent = successor;
                    }
                }
                this._transplant(nodeToDelete, successor);
                successor.left = nodeToDelete.left;
                if (this.isRealNode(successor.left)) {
                    successor.left.parent = successor;
                }
                successor.color = nodeToDelete.color;
            }
        }
        this._size--;
        // If the original color was black, fix the tree
        if (originalColor === 'BLACK') {
            this._deleteFixup(replacementNode);
        }
        results.push({ deleted: nodeToDelete, needBalanced: undefined });
        return results;
    }
    /**
     * The function sets the root of a tree-like structure and updates the parent property of the new
     * root.
     * @param {NODE | undefined} v - v is a parameter of type NODE or undefined.
     */
    _setRoot(v) {
        if (v) {
            v.parent = undefined;
        }
        this._root = v;
    }
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function replaces an old node with a new node while preserving the color of the old node.
     * @param {NODE} oldNode - The `oldNode` parameter represents the node that needs to be replaced in
     * the data structure.
     * @param {NODE} newNode - The `newNode` parameter is the new node that will replace the old node in
     * the data structure.
     * @returns The method is returning the result of calling the `_replaceNode` method from the
     * superclass, with the `oldNode` and `newNode` parameters.
     */
    _replaceNode(oldNode, newNode) {
        newNode.color = oldNode.color;
        return super._replaceNode(oldNode, newNode);
    }
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `_insert` function inserts or updates a node in a binary search tree and performs necessary
     * fix-ups to maintain the red-black tree properties.
     * @param {NODE} node - The `node` parameter represents the node that needs to be inserted into a
     * binary search tree. It contains a `key` property that is used to determine the position of the
     * node in the tree.
     * @returns {'inserted' | 'updated'} - The result of the insertion.
     */
    _insert(node) {
        var _a, _b;
        let current = this.root;
        let parent = undefined;
        while (this.isRealNode(current)) {
            parent = current;
            if (node.key < current.key) {
                current = (_a = current.left) !== null && _a !== void 0 ? _a : this.NIL;
            }
            else if (node.key > current.key) {
                current = (_b = current.right) !== null && _b !== void 0 ? _b : this.NIL;
            }
            else {
                this._replaceNode(current, node);
                return 'UPDATED';
            }
        }
        node.parent = parent;
        if (!parent) {
            this._setRoot(node);
        }
        else if (node.key < parent.key) {
            parent.left = node;
        }
        else {
            parent.right = node;
        }
        node.left = this.NIL;
        node.right = this.NIL;
        node.color = 'RED';
        this._insertFixup(node);
        return 'CREATED';
    }
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `_transplant` is used to replace a node `u` with another node `v` in a binary tree.
     * @param {NODE} u - The parameter "u" represents a node in a binary tree.
     * @param {NODE | undefined} v - The parameter `v` is of type `NODE | undefined`, which means it can
     * either be a `NODE` object or `undefined`.
     */
    _transplant(u, v) {
        if (!u.parent) {
            this._setRoot(v);
        }
        else if (u === u.parent.left) {
            u.parent.left = v;
        }
        else {
            u.parent.right = v;
        }
        if (v) {
            v.parent = u.parent;
        }
    }
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `_insertFixup` function is used to fix the Red-Black Tree after inserting a new node.
     * @param {NODE | undefined} z - The parameter `z` represents a node in the Red-Black Tree. It can
     * either be a valid node object or `undefined`.
     */
    _insertFixup(z) {
        var _a, _b, _c, _d;
        // Continue fixing the tree as long as the parent of z is red
        while (((_a = z === null || z === void 0 ? void 0 : z.parent) === null || _a === void 0 ? void 0 : _a.color) === 'RED') {
            // Check if the parent of z is the left child of its parent
            if (z.parent === ((_b = z.parent.parent) === null || _b === void 0 ? void 0 : _b.left)) {
                // Case 1: The uncle (y) of z is red
                const y = z.parent.parent.right;
                if ((y === null || y === void 0 ? void 0 : y.color) === 'RED') {
                    // Set colors to restore properties of Red-Black Tree
                    z.parent.color = 'BLACK';
                    y.color = 'BLACK';
                    z.parent.parent.color = 'RED';
                    // Move up the tree to continue fixing
                    z = z.parent.parent;
                }
                else {
                    // Case 2: The uncle (y) of z is black, and z is a right child
                    if (z === z.parent.right) {
                        // Perform a left rotation to transform the case into Case 3
                        z = z.parent;
                        this._leftRotate(z);
                    }
                    // Case 3: The uncle (y) of z is black, and z is a left child
                    // Adjust colors and perform a right rotation
                    if (z && this.isRealNode(z.parent) && this.isRealNode(z.parent.parent)) {
                        z.parent.color = 'BLACK';
                        z.parent.parent.color = 'RED';
                        this._rightRotate(z.parent.parent);
                    }
                }
            }
            else {
                // Symmetric case for the right child (left and right exchanged)
                // Follow the same logic as above with left and right exchanged
                const y = (_d = (_c = z === null || z === void 0 ? void 0 : z.parent) === null || _c === void 0 ? void 0 : _c.parent) === null || _d === void 0 ? void 0 : _d.left;
                if ((y === null || y === void 0 ? void 0 : y.color) === 'RED') {
                    z.parent.color = 'BLACK';
                    y.color = 'BLACK';
                    z.parent.parent.color = 'RED';
                    z = z.parent.parent;
                }
                else {
                    if (z === z.parent.left) {
                        z = z.parent;
                        this._rightRotate(z);
                    }
                    if (z && this.isRealNode(z.parent) && this.isRealNode(z.parent.parent)) {
                        z.parent.color = 'BLACK';
                        z.parent.parent.color = 'RED';
                        this._leftRotate(z.parent.parent);
                    }
                }
            }
        }
        // Ensure that the root is black after fixing
        if (this.isRealNode(this._root))
            this._root.color = 'BLACK';
    }
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(log n)
     * Space Complexity: O(1)
     *
     * The `_deleteFixup` function is used to fix the red-black tree after a node deletion by adjusting
     * the colors and performing rotations.
     * @param {NODE | undefined} node - The `node` parameter represents a node in a Red-Black Tree data
     * structure. It can be either a valid node object or `undefined`.
     * @returns The function does not return any value. It has a return type of `void`.
     */
    _deleteFixup(node) {
        var _a, _b, _c, _d;
        // Early exit condition
        if (!node || node === this.root || node.color === 'BLACK') {
            if (node) {
                node.color = 'BLACK'; // Ensure the final node is black
            }
            return;
        }
        while (node && node !== this.root && node.color === 'BLACK') {
            const parent = node.parent;
            if (!parent) {
                break; // Ensure the loop terminates if there's an issue with the tree structure
            }
            if (node === parent.left) {
                let sibling = parent.right;
                // Cases 1 and 2: Sibling is red or both children of sibling are black
                if ((sibling === null || sibling === void 0 ? void 0 : sibling.color) === 'RED') {
                    sibling.color = 'BLACK';
                    parent.color = 'RED';
                    this._leftRotate(parent);
                    sibling = parent.right;
                }
                // Case 3: Sibling's left child is black
                if (((_b = (_a = sibling === null || sibling === void 0 ? void 0 : sibling.left) === null || _a === void 0 ? void 0 : _a.color) !== null && _b !== void 0 ? _b : 'BLACK') === 'BLACK') {
                    if (sibling)
                        sibling.color = 'RED';
                    node = parent;
                }
                else {
                    // Case 4: Adjust colors and perform a right rotation
                    if (sibling === null || sibling === void 0 ? void 0 : sibling.left)
                        sibling.left.color = 'BLACK';
                    if (sibling)
                        sibling.color = parent.color;
                    parent.color = 'BLACK';
                    this._rightRotate(parent);
                    node = this.root;
                }
            }
            else {
                // Symmetric case for the right child (left and right exchanged)
                let sibling = parent.left;
                // Cases 1 and 2: Sibling is red or both children of sibling are black
                if ((sibling === null || sibling === void 0 ? void 0 : sibling.color) === 'RED') {
                    sibling.color = 'BLACK';
                    if (parent)
                        parent.color = 'RED';
                    this._rightRotate(parent);
                    if (parent)
                        sibling = parent.left;
                }
                // Case 3: Sibling's left child is black
                if (((_d = (_c = sibling === null || sibling === void 0 ? void 0 : sibling.right) === null || _c === void 0 ? void 0 : _c.color) !== null && _d !== void 0 ? _d : 'BLACK') === 'BLACK') {
                    if (sibling)
                        sibling.color = 'RED';
                    node = parent;
                }
                else {
                    // Case 4: Adjust colors and perform a left rotation
                    if (sibling === null || sibling === void 0 ? void 0 : sibling.right)
                        sibling.right.color = 'BLACK';
                    if (sibling)
                        sibling.color = parent.color;
                    if (parent)
                        parent.color = 'BLACK';
                    this._leftRotate(parent);
                    node = this.root;
                }
            }
        }
        // Ensure that the final node (possibly the root) is black
        if (node) {
            node.color = 'BLACK';
        }
    }
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_leftRotate` function performs a left rotation on a given node in a binary tree.
     * @param {NODE | undefined} x - The parameter `x` is of type `NODE | undefined`. It represents a
     * node in a binary tree or `undefined` if there is no node.
     * @returns void, which means it does not return any value.
     */
    _leftRotate(x) {
        if (!x || !x.right) {
            return;
        }
        const y = x.right;
        x.right = y.left;
        if (this.isRealNode(y.left)) {
            y.left.parent = x;
        }
        y.parent = x.parent;
        if (!x.parent) {
            this._setRoot(y);
        }
        else if (x === x.parent.left) {
            x.parent.left = y;
        }
        else {
            x.parent.right = y;
        }
        y.left = x;
        x.parent = y;
    }
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_rightRotate` function performs a right rotation on a given node in a binary tree.
     * @param {NODE | undefined} y - The parameter `y` is of type `NODE | undefined`. It represents a
     * node in a binary tree or `undefined` if there is no node.
     * @returns void, which means it does not return any value.
     */
    _rightRotate(y) {
        if (!y || !y.left) {
            return;
        }
        const x = y.left;
        y.left = x.right;
        if (this.isRealNode(x.right)) {
            x.right.parent = y;
        }
        x.parent = y.parent;
        if (!y.parent) {
            this._setRoot(x);
        }
        else if (y === y.parent.left) {
            y.parent.left = x;
        }
        else {
            y.parent.right = x;
        }
        x.right = y;
        y.parent = x;
    }
}
exports.RedBlackTree = RedBlackTree;
